from uuid import UUID

from django.urls import reverse
from rest_framework import status

from arkindex.project.tests import FixtureAPITestCase


class TestElementListsCentroid(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.element_type = cls.corpus.types.create(slug="potatoid", display_name="Potatoid", folder=False)
        cls.image = cls.imgsrv.images.get(path="img6")

        # We create some elements with centroids arranged on the image like so:
        # +------+
        # | A  B |
        # | C  D |
        # +------+
        # Two elements will have the same D centroid, so that we can test the fallback UUID sorting,
        # and two more elements will have no polygon so we can test that they are first or last, as well as their own UUID fallback.
        a = cls.corpus.elements.create(
            type=cls.element_type,
            name="A",
            image=cls.image,
            polygon=((10, 10), (10, 30), (30, 30), (30, 10), (10, 10)),
        )
        b = cls.corpus.elements.create(
            type=cls.element_type,
            name="B",
            image=cls.image,
            polygon=((110, 10), (110, 30), (130, 30), (130, 10), (110, 10)),
        )
        c = cls.corpus.elements.create(
            type=cls.element_type,
            name="C",
            image=cls.image,
            polygon=((10, 110), (10, 130), (30, 130), (30, 110), (10, 110)),
        )

        d1 = cls.corpus.elements.create(
            id=UUID("aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"),
            type=cls.element_type,
            name="D1",
            image=cls.image,
            polygon=((110, 110), (110, 130), (130, 130), (130, 110), (110, 110)),
        )
        d2 = cls.corpus.elements.create(
            id=UUID("aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaab"),
            type=cls.element_type,
            name="D2",
            image=cls.image,
            # Different polygon, but just 10 pixels larger on all sides, so it has the same centroid
            polygon=((100, 100), (100, 140), (140, 140), (140, 100), (100, 100)),
        )

        none1 = cls.corpus.elements.create(
            id=UUID("aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaa1"),
            type=cls.element_type,
            name="No polygon 1",
        )
        none2 = cls.corpus.elements.create(
            id=UUID("aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaa2"),
            type=cls.element_type,
            name="No polygon 2",
        )

        cls.elements = [a, b, c, d1, d2, none1, none2]

        cls.cases = [
            ["horizontal-lr", "asc", ["A", "B", "C", "D1", "D2", "No polygon 1", "No polygon 2"]],
            ["horizontal-rl", "asc", ["B", "A", "D1", "D2", "C", "No polygon 1", "No polygon 2"]],
            ["vertical-lr", "asc", ["A", "C", "B", "D1", "D2", "No polygon 1", "No polygon 2"]],
            ["vertical-rl", "asc", ["B", "D1", "D2", "A", "C", "No polygon 1", "No polygon 2"]],
            ["horizontal-lr", "desc", ["No polygon 2", "No polygon 1", "D2", "D1", "C", "B", "A"]],
            ["horizontal-rl", "desc", ["No polygon 2", "No polygon 1", "C", "D2", "D1", "A", "B"]],
            ["vertical-lr", "desc", ["No polygon 2", "No polygon 1", "D2", "D1", "B", "C", "A"]],
            ["vertical-rl", "desc", ["No polygon 2", "No polygon 1", "C", "A", "D2", "D1", "B"]],
        ]

    def test_list_elements(self):
        for order, order_direction, expected_elements in self.cases:
            with self.subTest(order=order, order_direction=order_direction):
                with self.assertNumQueries(6):
                    response = self.client.get(
                        reverse("api:corpus-elements", kwargs={"corpus": self.corpus.id}),
                        {"type": self.element_type.slug, "order": order, "order_direction": order_direction},
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)

                data = response.json()
                # Pagination is not supported in this test, so any response with more than 1 page could cause trouble.
                self.assertIsNone(data["next"])
                actual_elements = [element["name"] for element in response.json()["results"]]
                self.assertListEqual(expected_elements, actual_elements)

    def test_list_element_children(self):
        parent = self.corpus.elements.get(name="Volume 2")
        for element in self.elements:
            element.add_parent(parent)

        for order, order_direction, expected_elements in self.cases:
            with self.subTest(order=order, order_direction=order_direction):
                with self.assertNumQueries(6):
                    response = self.client.get(
                        reverse("api:elements-children", kwargs={"pk": str(parent.id)}),
                        {"type": self.element_type.slug, "order": order, "order_direction": order_direction},
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)

                data = response.json()
                # Pagination is not supported in this test, so any response with more than 1 page could cause trouble.
                self.assertIsNone(data["next"])
                actual_elements = [element["name"] for element in response.json()["results"]]
                self.assertListEqual(expected_elements, actual_elements)

    def test_list_element_parents(self):
        child = self.corpus.elements.get(name="Surface E")
        for element in self.elements:
            child.add_parent(element)

        for order, order_direction, expected_elements in self.cases:
            with self.subTest(order=order, order_direction=order_direction):
                with self.assertNumQueries(6):
                    response = self.client.get(
                        reverse("api:elements-parents", kwargs={"pk": str(child.id)}),
                        {"type": self.element_type.slug, "order": order, "order_direction": order_direction},
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)

                data = response.json()
                # Pagination is not supported in this test, so any response with more than 1 page could cause trouble.
                self.assertIsNone(data["next"])
                actual_elements = [element["name"] for element in response.json()["results"]]
                self.assertListEqual(expected_elements, actual_elements)
