from drf_spectacular.authentication import TokenScheme
from rest_framework.authentication import TokenAuthentication
from rest_framework.exceptions import AuthenticationFailed

from arkindex.ponos.models import Task
from sentry_sdk import set_tag


class TaskAuthentication(TokenAuthentication):
    keyword = "Ponos"
    model = Task

    def authenticate_credentials(self, key):
        try:
            task = Task.objects.select_related("process__creator", "process__corpus").get(token=key)
        except Task.DoesNotExist:
            # Same error message as the standard TokenAuthentication
            raise AuthenticationFailed("Invalid token.")

        # Add the process ID to Sentry tags, to speed up troubleshooting
        # by identifying if and which process is affected by the error.
        set_tag("process_id", str(task.process_id))

        if not task.process.creator_id or not task.process.creator.is_active:
            # Same error message as the standard TokenAuthentication
            raise AuthenticationFailed("User inactive or deleted.")

        # Must return a 2-tuple that will be set as (self.request.user, self.request.auth)
        return (task.process.creator, task)


class TaskAuthenticationExtension(TokenScheme):
    target_class = "arkindex.ponos.authentication.TaskAuthentication"
    name = "taskAuth"
    # The TokenScheme has a priority of -1 and matches both TokenAuthentication and its subclasses;
    # we set the priority to a higher number to make this extension match first, and disable
    # subclass matching so that this only applies to the TaskAuthentication.
    priority = 0
    match_subclasses = False
