from datetime import datetime, timezone
from unittest.mock import patch

from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, Element
from arkindex.ponos.models import Farm
from arkindex.process.models import (
    Process,
    ProcessMode,
    WorkerActivity,
    WorkerActivityState,
    WorkerConfiguration,
    WorkerVersion,
)
from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Model, ModelVersionState


@override_settings(PUBLIC_HOSTNAME="https://darkindex.lol")
class TestListWorkerActivities(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker_version = WorkerVersion.objects.get(worker__slug="reco")
        cls.element = Element.objects.get(name="Volume 1, page 2r")
        cls.process = Process.objects.create(
            mode=ProcessMode.Workers,
            creator=cls.user,
            corpus=cls.corpus,
            farm=Farm.objects.first(),
        )
        cls.configuration = WorkerConfiguration.objects.create(
            worker=cls.worker_version.worker,
            name="A config",
            configuration={"a": "b"},
        )
        cls.model = Model.objects.create(name="Mochi", public=False)
        cls.model_version = cls.model.versions.create(
            state=ModelVersionState.Available,
            hash="a",
        )

        cls.process.worker_runs.create(
            version=cls.worker_version,
            configuration=cls.configuration,
            model_version=cls.model_version,
        )

        # Run the process, but skip the real activity initialization so that we can control it ourselves
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            cls.process.run()

        # Create a queued activity for this element
        cls.activity = cls.element.activities.create(
            process=cls.process,
            worker_version=cls.worker_version,
            state=WorkerActivityState.Queued,
            configuration=cls.configuration,
            model_version=cls.model_version,
        )

    def test_list_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:corpus-activity", kwargs={"corpus": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:corpus-activity", kwargs={"corpus": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_list_corpus_acl(self, has_access_mock):
        private_corpus = Corpus.objects.create(name="can't touch this")
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:corpus-activity", kwargs={"corpus": private_corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have guest access to this corpus."})

    def test_list_corpus_not_found(self):
        corpus_id = "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"
        self.assertNotEqual(str(self.corpus.id), corpus_id)
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:corpus-activity", kwargs={"corpus": corpus_id}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_list(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:corpus-activity", kwargs={"corpus": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "count": 1,
            "number": 1,
            "previous": None,
            "next": None,
            "results": [
                {
                    "created": self.activity.created.isoformat().replace("+00:00", "Z"),
                    "updated": self.activity.updated.isoformat().replace("+00:00", "Z"),
                    "started": None,
                    "element_id": str(self.element.id),
                    "process_id": str(self.process.id),
                    "worker_version_id": str(self.worker_version.id),
                    "configuration_id": str(self.configuration.id),
                    "model_version_id": str(self.model_version.id),
                    "state": "queued"
                }
            ]
        })

    def test_list_started(self):
        self.activity.state = WorkerActivityState.Started
        self.activity.started = datetime(2012, 12, 12, 12, 12, 12, 12, timezone.utc)
        self.activity.save(update_fields=["state", "started"])
        # Fetch the updated `updated` timestamp from Postgres, since it is managed by a trigger
        self.activity.refresh_from_db()
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:corpus-activity", kwargs={"corpus": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "count": 1,
            "number": 1,
            "previous": None,
            "next": None,
            "results": [
                {
                    "created": self.activity.created.isoformat().replace("+00:00", "Z"),
                    "updated": self.activity.updated.isoformat().replace("+00:00", "Z"),
                    "started": "2012-12-12T12:12:12.000012Z",
                    "element_id": str(self.element.id),
                    "process_id": str(self.process.id),
                    "worker_version_id": str(self.worker_version.id),
                    "configuration_id": str(self.configuration.id),
                    "model_version_id": str(self.model_version.id),
                    "state": "started",
                }
            ]
        })

    def test_list_filter_process_wrong_corpus(self):
        private_corpus = Corpus.objects.create(name="can't touch this")
        process = private_corpus.processes.create(
            mode=ProcessMode.Workers,
            creator=self.superuser,
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.get(
                reverse("api:corpus-activity", kwargs={"corpus": self.corpus.id}),
                {"process_id": str(process.id)}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"process_id": ["This process does not exist."]})

    def test_list_filters(self):
        # Delete all worker activities
        WorkerActivity.objects.all().delete()

        # Create objects
        process_1 = self.corpus.processes.create(
            mode=ProcessMode.Workers,
            creator=self.user,
        )
        process_2 = self.corpus.processes.create(
            mode=ProcessMode.Workers,
            creator=self.user,
        )
        worker_version_2 = WorkerVersion.objects.get(worker__slug="dla")
        model_version_2 = self.model.versions.create(
            state=ModelVersionState.Available,
            hash="b",
        )
        configuration_2 = WorkerConfiguration.objects.create(worker=self.worker_version.worker, name="Another configuration", configuration={"c": "d"})

        # Create all the worker activities
        activity_1 = process_1.activities.create(
            element=self.element,
            worker_version=self.worker_version,
            state=WorkerActivityState.Processed,
            model_version=self.model_version,
            configuration=self.configuration
        )
        activity_2 = process_1.activities.create(
            element=self.element,
            worker_version=worker_version_2,
            state=WorkerActivityState.Error,
            model_version=self.model_version,
            configuration=self.configuration
        )
        activity_3 = process_1.activities.create(
            element=self.element,
            worker_version=self.worker_version,
            state=WorkerActivityState.Processed,
            model_version=self.model_version,
            configuration=configuration_2
        )
        activity_4 = process_1.activities.create(
            element=self.element,
            worker_version=self.worker_version,
            state=WorkerActivityState.Processed,
            model_version=model_version_2,
            configuration=configuration_2
        )
        activity_5 = process_2.activities.create(
            element=self.element,
            worker_version=worker_version_2,
            state=WorkerActivityState.Processed,
            model_version=model_version_2,
            configuration=self.configuration
        )
        activity_6 = process_1.activities.create(
            element=self.element,
            worker_version=self.worker_version,
            state=WorkerActivityState.Processed,
            model_version=self.model_version,
            configuration=None
        )

        # Worker activity responses
        worker_activities = {
            activity.id: {
                "created": activity.created.isoformat().replace("+00:00", "Z"),
                "updated": activity.updated.isoformat().replace("+00:00", "Z"),
                "started": None,
                "element_id": str(self.element.id),
                "process_id": str(activity.process_id),
                "worker_version_id": str(activity.worker_version.id),
                "configuration_id": str(activity.configuration_id) if activity.configuration_id else None,
                "model_version_id": str(activity.model_version_id) if activity.model_version_id else None,
                "state": activity.state.value
            }
            for activity in WorkerActivity.objects.all()
        }

        # Test cases
        cases = [
            (
                {"process_id": str(process_1.id)}, [activity_1.id, activity_2.id, activity_3.id, activity_4.id, activity_6.id]
            ),
            (
                {"state": "processed"}, [activity_1.id, activity_3.id, activity_4.id, activity_5.id, activity_6.id]
            ),
            (
                {"process_id": str(process_1.id), "state": "error"}, [activity_2.id]
            ),
            (
                {"process_id": str(process_1.id), "state": "processed", "worker_version_id": str(self.worker_version.id)},
                [activity_1.id, activity_3.id, activity_4.id, activity_6.id]
            ),
            (
                {"worker_version_id": str(worker_version_2.id)}, [activity_2.id, activity_5.id]
            ),
            (
                {
                    "process_id": str(process_1.id),
                    "state": "processed",
                    "worker_version_id": str(self.worker_version.id),
                    "model_version_id": str(self.model_version.id)
                },
                [activity_1.id, activity_3.id, activity_6.id]
            ),
            (
                {"model_version_id": str(model_version_2.id)}, [activity_4.id, activity_5.id]
            ),
            (
                {
                    "process_id": str(process_1.id),
                    "state": "processed",
                    "worker_version_id": str(self.worker_version.id),
                    "model_version_id": str(self.model_version.id),
                    "worker_configuration_id": str(self.configuration.id)
                },
                [activity_1.id]
            ),
            (
                {"worker_configuration_id": str(configuration_2.id)}, [activity_3.id, activity_4.id]
            ),
            (
                {
                    "process_id": str(process_1.id),
                    "state": "processed",
                    "worker_version_id": str(self.worker_version.id),
                    "model_version_id": False,
                    "worker_configuration_id": str(self.configuration.id)
                },
                []
            ),
            (
                {
                    "process_id": str(process_1.id),
                    "state": "processed",
                    "worker_version_id": str(self.worker_version.id),
                    "model_version_id": str(self.model_version.id),
                    "worker_configuration_id": False
                },
                [activity_6.id]
            ),
        ]

        self.client.force_login(self.user)
        for filters, activity_ids in cases:
            # Sort activities by ID, like in the API response
            activity_ids.sort()

            # Filtering by process_id adds 1 query as it checks if the process exists
            queries_count = 6 if "process_id" in filters else 5
            # If there are no results returned, only the request to return a 'count' is made so we get one less query
            if not len(activity_ids):
                queries_count -= 1
            with self.assertNumQueries(queries_count):
                response = self.client.get(
                    reverse("api:corpus-activity", kwargs={"corpus": self.corpus.id}),
                    {**filters}
                )
                self.assertEqual(response.status_code, status.HTTP_200_OK)

            self.assertDictEqual(response.json(), {
                "count": len(activity_ids),
                "number": 1,
                "previous": None,
                "next": None,
                "results": [
                    worker_activities[id] for id in activity_ids
                ]
            })

    def test_list_invalid_filters(self):
        self.client.force_login(self.superuser)
        cases = [
            (
                {"process_id": "a", "worker_version_id": "neon", "model_version_id": "genesis", "worker_configuration_id": "evangelion"},
                {
                    "process_id": ["Not a valid UUID."],
                    "worker_version_id": ["Not a valid UUID."],
                    "model_version_id": ["Not a valid UUID."],
                    "worker_configuration_id": ["Not a valid UUID."]
                },
            ),
            (
                {"process_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"},
                {"process_id": ["This process does not exist."]},
            ),
            (
                {"state": "lol"},
                {"state": ["'lol' is not a valid WorkerActivity state"]},
            ),
            (
                {"process_id": "a", "state": "lol"},
                {
                    "process_id": ["Not a valid UUID."],
                    "state": ["'lol' is not a valid WorkerActivity state"],
                },
            ),
        ]
        for filters, expected_errors in cases:
            with self.subTest(**filters):
                response = self.client.get(
                    reverse("api:corpus-activity", kwargs={"corpus": self.corpus.id}),
                    filters
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), expected_errors)
